@@ -1,5 +1,4 @@
-// -*- C++ -*-
-/* Copyright (C) 1989-2018 Free Software Foundation, Inc.
+/* Copyright (C) 1989-2021 Free Software Foundation, Inc.
      Written by James Clark (jjc@jclark.com)
 
 This file is part of groff.
@@ -19,11 +18,12 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include "lib.h"
 
-#include <ctype.h>
 #include <assert.h>
+#include <ctype.h>
 #include <math.h>
 #include <stdlib.h>
 #include <wchar.h>
+
 #include "errarg.h"
 #include "error.h"
 #include "cset.h"
@@ -69,75 +69,77 @@ struct text_file {
   FILE *fp;
   char *path;
   int lineno;
-  int size;
-  int skip_comments;
-  int silent;
+  int linebufsize;
+  bool recognize_comments;
+  bool silent;
   char *buf;
   text_file(FILE *fp, char *p);
   ~text_file();
-  int next();
-  void error(const char *format, 
+  bool next_line();
+  void error(const char *format,
+	     const errarg &arg1 = empty_errarg,
+	     const errarg &arg2 = empty_errarg,
+	     const errarg &arg3 = empty_errarg);
+  void fatal(const char *format,
 	     const errarg &arg1 = empty_errarg,
 	     const errarg &arg2 = empty_errarg,
 	     const errarg &arg3 = empty_errarg);
 };
 
-text_file::text_file(FILE *p, char *s) 
-: fp(p), path(s), lineno(0), size(0), skip_comments(1), silent(0), buf(0)
+text_file::text_file(FILE *p, char *s) : fp(p), path(s), lineno(0),
+  linebufsize(128), recognize_comments(true), silent(false), buf(0)
 {
 }
 
 text_file::~text_file()
 {
-  a_delete buf;
+  delete[] buf;
   free(path);
   if (fp)
     fclose(fp);
 }
 
-int text_file::next()
+bool text_file::next_line()
 {
   if (fp == 0)
-    return 0;
-  if (buf == 0) {
-    buf = new char[128];
-    size = 128;
-  }
+    return false;
+  if (buf == 0)
+    buf = new char[linebufsize];
   for (;;) {
-    int i = 0;
+    lineno++;
+    int length = 0;
     for (;;) {
       int c = getc(fp);
       if (c == EOF)
 	break;
-      if (invalid_input_char(c))
-	error("invalid input character code '%1'", int(c));
+      if (is_invalid_input_char(c))
+	error("invalid input character code %1", int(c));
       else {
-	if (i + 1 >= size) {
+	if (length + 1 >= linebufsize) {
 	  char *old_buf = buf;
-	  buf = new char[size*2];
-	  memcpy(buf, old_buf, size);
-	  a_delete old_buf;
-	  size *= 2;
+	  buf = new char[linebufsize * 2];
+	  memcpy(buf, old_buf, linebufsize);
+	  delete[] old_buf;
+	  linebufsize *= 2;
 	}
-	buf[i++] = c;
+	buf[length++] = c;
 	if (c == '\n')
 	  break;
       }
     }
-    if (i == 0)
+    if (length == 0)
       break;
-    buf[i] = '\0';
-    lineno++;
+    buf[length] = '\0';
     char *ptr = buf;
     while (csspace(*ptr))
       ptr++;
-    if (*ptr != 0 && (!skip_comments || *ptr != '#'))
-      return 1;
+    if (*ptr != 0 && (!recognize_comments || *ptr != '#'))
+      return true;
   }
-  return 0;
+  return false;
 }
 
-void text_file::error(const char *format, 
+void text_file::error(const char *format,
 		      const errarg &arg1,
 		      const errarg &arg2,
 		      const errarg &arg3)
@@ -146,10 +148,19 @@ void text_file::error(const char *format,
     error_with_file_and_line(path, lineno, format, arg1, arg2, arg3);
 }
 
+void text_file::fatal(const char *format,
+		      const errarg &arg1,
+		      const errarg &arg2,
+		      const errarg &arg3)
+{
+  if (!silent)
+    fatal_with_file_and_line(path, lineno, format, arg1, arg2, arg3);
+}
+
 int glyph_to_unicode(glyph *g)
 {
   const char *nm = glyph_to_name(g);
-  if (nm != NULL) {
+  if (nm != 0) {
     // ASCII character?
     if (nm[0] == 'c' && nm[1] == 'h' && nm[2] == 'a' && nm[3] == 'r'
 	&& (nm[4] >= '0' && nm[4] <= '9')) {
@@ -180,7 +191,7 @@ int glyph_to_unicode(glyph *g)
     }
     // groff glyphs that map to Unicode?
     const char *unicode = glyph_name_to_unicode(nm);
-    if (unicode != NULL && strchr(unicode, '_') == NULL) {
+    if (unicode != 0 && strchr(unicode, '_') == 0) {
       char *ignore;
       return (int)strtol(unicode, &ignore, 16);
     }
@@ -190,25 +201,22 @@ int glyph_to_unicode(glyph *g)
 
 /* font functions */
 
-font::font(const char *s)
-: ligatures(0), kern_hash_table(0), space_width(0), special(0),
-  ch_index(0), nindices(0), ch(0), ch_used(0), ch_size(0), widths_cache(0)
+font::font(const char *s) : ligatures(0), kern_hash_table(0),
+  space_width(0), special(false), internalname(0), slant(0.0), zoom(0),
+  ch_index(0), nindices(0), ch(0), ch_used(0), ch_size(0),
+  widths_cache(0)
 {
   name = new char[strlen(s) + 1];
   strcpy(name, s);
-  internalname = 0;
-  slant = 0.0;
-  zoom = 0;
-  // load();			// for testing
 }
 
 font::~font()
 {
   for (int i = 0; i < ch_used; i++)
     if (ch[i].special_device_coding)
-      a_delete ch[i].special_device_coding;
-  a_delete ch;
-  a_delete ch_index;
+      delete[] ch[i].special_device_coding;
+  delete[] ch;
+  delete[] ch_index;
   if (kern_hash_table) {
     for (int i = 0; i < KERN_HASH_TABLE_SIZE; i++) {
       font_kern_list *kerns = kern_hash_table[i];
@@ -218,10 +226,10 @@ font::~font()
 	delete tem;
       }
     }
-    a_delete kern_hash_table;
+    delete[] kern_hash_table;
   }
-  a_delete name;
-  a_delete internalname;
+  delete[] name;
+  delete[] internalname;
   while (widths_cache) {
     font_widths_cache *tem = widths_cache;
     widths_cache = widths_cache->next;
@@ -266,9 +274,11 @@ inline int font::scale(int w, int sz)
     return sz == unitwidth ? w : scale_round(w, sz, unitwidth);
 }
 
-int font::unit_scale(double *value, char unit)
+// Returns whether scaling by arguments was successful.  Used for paper
+// size conversions.
+bool font::unit_scale(double *value, char unit)
 {
-  // we scale everything to inch
+  // Paper sizes are handled in inches.
   double divisor = 0;
   switch (unit) {
   case 'i':
@@ -284,14 +294,14 @@ int font::unit_scale(double *value, char unit)
     divisor = 2.54;
     break;
   default:
-    assert(0);
+    assert(0 == "unit not in [cipP]");
     break;
   }
   if (divisor) {
     *value /= divisor;
-    return 1;
+    return true;
   }
-  return 0;
+  return false;
 }
 
 int font::get_skew(glyph *g, int point_size, int sl)
@@ -300,26 +310,26 @@ int font::get_skew(glyph *g, int point_size, int sl)
   return int(h * tan((slant + sl) * PI / 180.0) + .5);
 }
 
-int font::contains(glyph *g)
+bool font::contains(glyph *g)
 {
   int idx = glyph_to_index(g);
   assert(idx >= 0);
   // Explicitly enumerated glyph?
   if (idx < nindices && ch_index[idx] >= 0)
-    return 1;
+    return true;
   if (is_unicode) {
     // Unicode font
     // ASCII or Unicode character, or groff glyph name that maps to Unicode?
     if (glyph_to_unicode(g) >= 0)
-      return 1;
+      return true;
     // Numbered character?
     if (glyph_to_number(g) >= 0)
-      return 1;
+      return true;
   }
-  return 0;
+  return false;
 }
 
-int font::is_special()
+bool font::is_special()
 {
   return special;
 }
@@ -335,7 +345,7 @@ font_widths_cache::font_widths_cache(int ps, int ch_size,
 
 font_widths_cache::~font_widths_cache()
 {
-  a_delete width;
+  delete[] width;
 }
 
 int font::get_width(glyph *g, int point_size)
@@ -343,7 +353,7 @@ int font::get_width(glyph *g, int point_size)
   int idx = glyph_to_index(g);
   assert(idx >= 0);
   int real_size;
-  if (!zoom)
+  if (zoom == 0) // 0 means "don't zoom"
     real_size = point_size;
   else
   {
@@ -355,7 +365,7 @@ int font::get_width(glyph *g, int point_size)
   if (idx < nindices && ch_index[idx] >= 0) {
     // Explicitly enumerated glyph
     int i = ch_index[idx];
-    if (real_size == unitwidth || font::unscaled_charwidths)
+    if (real_size == unitwidth || font::use_unscaled_charwidths)
       return ch[i].width;
 
     if (!widths_cache)
@@ -386,12 +396,13 @@ int font::get_width(glyph *g, int point_size)
     int w = wcwidth(get_code(g));
     if (w > 1)
       width *= w;
-    if (real_size == unitwidth || font::unscaled_charwidths)
+    if (real_size == unitwidth || font::use_unscaled_charwidths)
       return width;
     else
       return scale(width, point_size);
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 int font::get_height(glyph *g, int point_size)
@@ -406,7 +417,8 @@ int font::get_height(glyph *g, int point_size)
     // Unicode font
     return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 int font::get_depth(glyph *g, int point_size)
@@ -421,7 +433,8 @@ int font::get_depth(glyph *g, int point_size)
     // Unicode font
     return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 int font::get_italic_correction(glyph *g, int point_size)
@@ -436,7 +449,8 @@ int font::get_italic_correction(glyph *g, int point_size)
     // Unicode font
     return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 int font::get_left_italic_correction(glyph *g, int point_size)
@@ -451,7 +465,8 @@ int font::get_left_italic_correction(glyph *g, int point_size)
     // Unicode font
     return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 int font::get_subscript_correction(glyph *g, int point_size)
@@ -466,7 +481,8 @@ int font::get_subscript_correction(glyph *g, int point_size)
     // Unicode font
     return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 void font::set_zoom(int factor)
@@ -522,9 +538,9 @@ int font::get_kern(glyph *g1, glyph *g2, int point_size)
   return 0;
 }
 
-int font::has_ligature(int mask)
+bool font::has_ligature(int mask)
 {
-  return mask & ligatures;
+  return (bool) (mask & ligatures);
 }
 
 int font::get_character_type(glyph *g)
@@ -539,7 +555,8 @@ int font::get_character_type(glyph *g)
     // Unicode font
     return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 int font::get_code(glyph *g)
@@ -562,7 +579,8 @@ int font::get_code(glyph *g)
       return n;
   }
   // The caller must check 'contains(g)' before calling get_code(g).
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 const char *font::get_name()
@@ -585,9 +603,10 @@ const char *font::get_special_device_encoding(glyph *g)
   }
   if (is_unicode) {
     // Unicode font
-    return NULL;
+    return 0;
   }
-  abort();
+  assert(0 == "glyph is not indexed and device lacks Unicode support");
+  abort(); // -Wreturn-type
 }
 
 const char *font::get_image_generator()
@@ -612,10 +631,10 @@ void font::alloc_ch_index(int idx)
       nindices = idx + 10;
     int *old_ch_index = ch_index;
     ch_index = new int[nindices];
-    memcpy(ch_index, old_ch_index, sizeof(int)*old_nindices);
+    memcpy(ch_index, old_ch_index, sizeof(int) * old_nindices);
     for (int i = old_nindices; i < nindices; i++)
       ch_index[i] = -1;
-    a_delete old_ch_index;
+    delete[] old_ch_index;
   }
 }
 
@@ -628,8 +647,8 @@ void font::extend_ch()
     ch_size *= 2;
     font_char_metric *old_ch = ch;
     ch = new font_char_metric[ch_size];
-    memcpy(ch, old_ch, old_ch_size*sizeof(font_char_metric));
-    a_delete old_ch;
+    memcpy(ch, old_ch, old_ch_size * sizeof(font_char_metric));
+    delete[] old_ch;
   }
 }
 
@@ -644,14 +663,14 @@ void font::compact()
     int *old_ch_index = ch_index;
     ch_index = new int[i];
     memcpy(ch_index, old_ch_index, i*sizeof(int));
-    a_delete old_ch_index;
+    delete[] old_ch_index;
     nindices = i;
   }
   if (ch_used < ch_size) {
     font_char_metric *old_ch = ch;
     ch = new font_char_metric[ch_used];
     memcpy(ch, old_ch, ch_used*sizeof(font_char_metric));
-    a_delete old_ch;
+    delete[] old_ch;
     ch_size = ch_used;
   }
 }
@@ -680,10 +699,10 @@ void font::copy_entry(glyph *new_glyph, glyph *old_glyph)
   ch_index[new_index] = ch_index[old_index];
 }
 
-font *font::load_font(const char *s, int *not_found, int head_only)
+font *font::load_font(const char *s, bool load_header_only)
 {
   font *f = new font(s);
-  if (!f->load(not_found, head_only)) {
+  if (!f->load(load_header_only)) {
     delete f;
     return 0;
   }
@@ -692,7 +711,7 @@ font *font::load_font(const char *s, int *not_found, int head_only)
 
 static char *trim_arg(char *p)
 {
-  if (!p)
+  if (0 == p)
     return 0;
   while (csspace(*p))
     p++;
@@ -703,13 +722,13 @@ static char *trim_arg(char *p)
   return p;
 }
 
-int font::scan_papersize(const char *p,
-			 const char **size, double *length, double *width)
+bool font::scan_papersize(const char *p, const char **size,
+			  double *length, double *width)
 {
   double l, w;
   char lu[2], wu[2];
   const char *pp = p;
-  int test_file = 1;
+  bool attempt_file_open = true;
   char line[255];
 again:
   if (csdigit(*pp)) {
@@ -722,7 +741,7 @@ again:
 	*width = w;
       if (size)
 	*size = "custom";
-      return 1;
+      return true;
     }
   }
   else {
@@ -735,81 +754,91 @@ again:
 	  *width = papersizes[i].width;
 	if (size)
 	  *size = papersizes[i].name;
-	return 1;
+	return true;
       }
-    if (test_file) {
-      FILE *f = fopen(p, "r");
-      if (f) {
-	fgets(line, 254, f);
-	fclose(f);
-	test_file = 0;
-	char *linep = strchr(line, '\0');
-	// skip final newline, if any
-	if (*(--linep) == '\n')
-	  *linep = '\0';
-	pp = line;
+    if (attempt_file_open) {
+      FILE *fp = fopen(p, "r");
+      if (fp != 0) {
+	if (fgets(line, 254, fp)) {
+	  // Don't recurse on file names.
+	  attempt_file_open = false;
+	  char *linep = strchr(line, '\0');
+	  // skip final newline, if any
+	  if (*(--linep) == '\n')
+	    *linep = '\0';
+	  pp = line;
+	}
+	fclose(fp);
 	goto again;
       }
     }
   }
-  return 0;
+  return false;
 }
 
-// If the font can't be found, then if not_found is non-NULL, it will be set
-// to 1 otherwise a message will be printed.
-
-int font::load(int *not_found, int head_only)
+bool font::load(bool load_header_only)
 {
-  if (strcmp(name, "DESC") == 0) {
-    if (not_found)
-      *not_found = 1;
-    else
-      error("'DESC' is not a valid font file name");
-    return 0;
-  }
-  char *path;
   FILE *fp;
-  if ((fp = open_file(name, &path)) == NULL) {
-    if (not_found)
-      *not_found = 1;
-    else
-      error("can't find font file '%1'", name);
-    return 0;
-  }
+  char *path;
+  if ((fp = open_file(name, &path)) == 0)
+    return false;
   text_file t(fp, path);
-  t.skip_comments = 1;
-  t.silent = head_only;
-  char *p;
-  for (;;) {
-    if (!t.next()) {
-      p = 0;
-      break;
-    }
+  t.silent = load_header_only;
+  char *p = 0;
+  bool saw_name_directive = false;
+  while (t.next_line()) {
     p = strtok(t.buf, WS);
     if (strcmp(p, "name") == 0) {
+      p = strtok(0, WS);
+      if (0 == p) {
+	t.error("'name' directive requires an argument");
+	return false;
+      }
+      if (strcmp(p, name) != 0) {
+	t.error("font description file name '%1' does not match 'name'"
+		" argument '%2'", name, p);
+	return false;
+      }
+      saw_name_directive = true;
     }
     else if (strcmp(p, "spacewidth") == 0) {
       p = strtok(0, WS);
       int n;
-      if (p == 0 || sscanf(p, "%d", &n) != 1 || n <= 0) {
-	t.error("bad argument for 'spacewidth' command");
-	return 0;
+      if (0 == p) {
+	t.error("missing argument to 'spacewidth' directive");
+	return false;
+      }
+      if (sscanf(p, "%d", &n) != 1) {
+	t.error("invalid argument '%1' to 'spacewidth' directive", p);
+	return false;
+      }
+      if (n <= 0) {
+	t.error("'spacewidth' argument '%1' out of range", n);
+	return false;
       }
       space_width = n;
     }
     else if (strcmp(p, "slant") == 0) {
       p = strtok(0, WS);
       double n;
-      if (p == 0 || sscanf(p, "%lf", &n) != 1 || n >= 90.0 || n <= -90.0) {
-	t.error("bad argument for 'slant' command", p);
-	return 0;
+      if (0 == p) {
+	t.error("missing argument to 'slant' directive");
+	return false;
+      }
+      if (sscanf(p, "%lf", &n) != 1) {
+	t.error("invalid argument '%1' to 'slant' directive", p);
+	return false;
+      }
+      if (n >= 90.0 || n <= -90.0) {
+	t.error("'slant' directive argument '%1' out of range", n);
+	return false;
       }
       slant = n;
     }
     else if (strcmp(p, "ligatures") == 0) {
       for (;;) {
 	p = strtok(0, WS);
-	if (p == 0 || strcmp(p, "0") == 0)
+	if (0 == p || strcmp(p, "0") == 0)
 	  break;
 	if (strcmp(p, "ff") == 0)
 	  ligatures |= LIG_ff;
@@ -822,198 +851,203 @@ int font::load(int *not_found, int head_only)
 	else if (strcmp(p, "ffl") == 0)
 	  ligatures |= LIG_ffl;
 	else {
-	  t.error("unrecognised ligature '%1'", p);
-	  return 0;
+	  t.error("unrecognized ligature '%1'", p);
+	  return false;
 	}
       }
     }
     else if (strcmp(p, "internalname") == 0) {
       p = strtok(0, WS);
-      if (!p) {
-	t.error("'internalname' command requires argument");
-	return 0;
+      if (0 == p) {
+	t.error("missing argument to 'internalname' directive");
+	return false;
       }
       internalname = new char[strlen(p) + 1];
       strcpy(internalname, p);
     }
     else if (strcmp(p, "special") == 0) {
-      special = 1;
+      special = true;
     }
     else if (strcmp(p, "kernpairs") != 0 && strcmp(p, "charset") != 0) {
-      char *command = p;
+      char *directive = p;
       p = strtok(0, "\n");
-      handle_unknown_font_command(command, trim_arg(p), t.path, t.lineno);
+      handle_unknown_font_command(directive, trim_arg(p), t.path,
+				  t.lineno);
     }
     else
       break;
   }
-  int had_charset = 0;
-  if (p == 0) {
-    if (!is_unicode) {
-      t.error("missing charset command");
-      return 0;
-    }
-  } else {
-    char *command = p;
-    t.skip_comments = 0;
-    while (command) {
-      if (strcmp(command, "kernpairs") == 0) {
-	if (head_only)
-	  return 1;
-	for (;;) {
-	  if (!t.next()) {
-	    command = 0;
-	    break;
-	  }
-	  char *c1 = strtok(t.buf, WS);
-	  if (c1 == 0)
-	    continue;
-	  char *c2 = strtok(0, WS);
-	  if (c2 == 0) {
-	    command = c1;
-	    break;
-	  }
-	  p = strtok(0, WS);
-	  if (p == 0) {
-	    t.error("missing kern amount");
-	    return 0;
-	  }
-	  int n;
-	  if (sscanf(p, "%d", &n) != 1) {
-	    t.error("bad kern amount '%1'", p);
-	    return 0;
-	  }
-	  glyph *g1 = name_to_glyph(c1);
-	  glyph *g2 = name_to_glyph(c2);
-	  add_kern(g1, g2, n);
+  bool saw_charset_directive = false;
+  char *directive = p;
+  t.recognize_comments = false;
+  while (directive) {
+    if (strcmp(directive, "kernpairs") == 0) {
+      if (load_header_only)
+	return true;
+      for (;;) {
+	if (!t.next_line()) {
+	  directive = 0;
+	  break;
 	}
+	char *c1 = strtok(t.buf, WS);
+	if (0 == c1)
+	  continue;
+	char *c2 = strtok(0, WS);
+	if (0 == c2) {
+	  directive = c1;
+	  break;
+	}
+	p = strtok(0, WS);
+	if (0 == p) {
+	  t.error("missing kern amount for kerning pair '%1 %2'", c1,
+		  c2);
+	  return false;
+	}
+	int n;
+	if (sscanf(p, "%d", &n) != 1) {
+	  t.error("invalid kern amount '%1' for kerning pair '%2 %3'",
+		  p, c1, c2);
+	  return false;
+	}
+	glyph *g1 = name_to_glyph(c1);
+	glyph *g2 = name_to_glyph(c2);
+	add_kern(g1, g2, n);
       }
-      else if (strcmp(command, "charset") == 0) {
-	if (head_only)
-	  return 1;
-	had_charset = 1;
-	glyph *last_glyph = NULL;
-	for (;;) {
-	  if (!t.next()) {
-	    command = 0;
-	    break;
+    }
+    else if (strcmp(directive, "charset") == 0) {
+      if (load_header_only)
+	return true;
+      saw_charset_directive = true;
+      glyph *last_glyph = 0;
+      for (;;) {
+	if (!t.next_line()) {
+	  directive = 0;
+	  break;
+	}
+	char *nm = strtok(t.buf, WS);
+	assert(nm != 0);
+	p = strtok(0, WS);
+	if (0 == p) {
+	  directive = nm;
+	  break;
+	}
+	if (p[0] == '"') {
+	  if (last_glyph == 0) {
+	    t.error("the first entry ('%1') in 'charset' subsection"
+		    " cannot be an alias", nm);
+	    return false;
+	  }
+	  if (strcmp(nm, "---") == 0) {
+	    t.error("an unnamed character ('---') cannot be an alias");
+	    return false;
+	  }
+	  glyph *g = name_to_glyph(nm);
+	  copy_entry(g, last_glyph);
+	}
+	else {
+	  font_char_metric metric;
+	  metric.height = 0;
+	  metric.depth = 0;
+	  metric.pre_math_space = 0;
+	  metric.italic_correction = 0;
+	  metric.subscript_correction = 0;
+	  int nparms = sscanf(p, "%d,%d,%d,%d,%d,%d",
+			      &metric.width, &metric.height,
+			      &metric.depth,
+			      &metric.italic_correction,
+			      &metric.pre_math_space,
+			      &metric.subscript_correction);
+	  if (nparms < 1) {
+	    t.error("missing or invalid width for glyph '%1'", nm);
+	    return false;
 	  }
-	  char *nm = strtok(t.buf, WS);
-	  if (nm == 0)
-	    continue;			// I dont think this should happen
 	  p = strtok(0, WS);
-	  if (p == 0) {
-	    command = nm;
-	    break;
+	  if (0 == p) {
+	    t.error("missing character type for '%1'", nm);
+	    return false;
 	  }
-	  if (p[0] == '"') {
-	    if (last_glyph == NULL) {
-	      t.error("first charset entry is duplicate");
-	      return 0;
-	    }
-	    if (strcmp(nm, "---") == 0) {
-	      t.error("unnamed character cannot be duplicate");
-	      return 0;
-	    }
-	    glyph *g = name_to_glyph(nm);
-	    copy_entry(g, last_glyph);
+	  int type;
+	  if (sscanf(p, "%d", &type) != 1) {
+	    t.error("invalid character type for '%1'", nm);
+	    return false;
+	  }
+	  if (type < 0 || type > 255) {
+	    t.error("character type '%1' out of range for '%2'", type,
+		    nm);
+	    return false;
+	  }
+	  metric.type = type;
+	  p = strtok(0, WS);
+	  if (0 == p) {
+	    t.error("missing code for '%1'", nm);
+	    return false;
+	  }
+	  char *ptr;
+	  metric.code = (int)strtol(p, &ptr, 0);
+	  if (metric.code == 0 && ptr == p) {
+	    t.error("invalid code '%1' for character '%2'", p, nm);
+	    return false;
+	  }
+	  if (is_unicode) {
+	    int w = wcwidth(metric.code);
+	    if (w > 1)
+	      metric.width *= w;
+	  }
+	  p = strtok(0, WS);
+	  if ((0 == p) || (strcmp(p, "--") == 0)) {
+	    metric.special_device_coding = 0;
 	  }
 	  else {
-	    font_char_metric metric;
-	    metric.height = 0;
-	    metric.depth = 0;
-	    metric.pre_math_space = 0;
-	    metric.italic_correction = 0;
-	    metric.subscript_correction = 0;
-	    int nparms = sscanf(p, "%d,%d,%d,%d,%d,%d",
-				&metric.width, &metric.height, &metric.depth,
-				&metric.italic_correction,
-				&metric.pre_math_space,
-				&metric.subscript_correction);
-	    if (nparms < 1) {
-	      t.error("bad width for '%1'", nm);
-	      return 0;
-	    }
-	    p = strtok(0, WS);
-	    if (p == 0) {
-	      t.error("missing character type for '%1'", nm);
-	      return 0;
-	    }
-	    int type;
-	    if (sscanf(p, "%d", &type) != 1) {
-	      t.error("bad character type for '%1'", nm);
-	      return 0;
-	    }
-	    if (type < 0 || type > 255) {
-	      t.error("character type '%1' out of range", type);
-	      return 0;
-	    }
-	    metric.type = type;
-	    p = strtok(0, WS);
-	    if (p == 0) {
-	      t.error("missing code for '%1'", nm);
-	      return 0;
-	    }
-	    char *ptr;
-	    metric.code = (int)strtol(p, &ptr, 0);
-	    if (metric.code == 0 && ptr == p) {
-	      t.error("bad code '%1' for character '%2'", p, nm);
-	      return 0;
-	    }
-	    if (is_unicode) {
-	      int w = wcwidth(metric.code);
-	      if (w > 1)
-	        metric.width *= w;
-	    }
-	    p = strtok(0, WS);
-	    if ((p == NULL) || (strcmp(p, "--") == 0)) {
-	      metric.special_device_coding = NULL;
-	    }
-	    else {
-	      char *nam = new char[strlen(p) + 1];
-	      strcpy(nam, p);
-	      metric.special_device_coding = nam;
-	    }
-	    if (strcmp(nm, "---") == 0) {
-	      last_glyph = number_to_glyph(metric.code);
-	      add_entry(last_glyph, metric);
-	    }
-	    else {
-	      last_glyph = name_to_glyph(nm);
-	      add_entry(last_glyph, metric);
-	      copy_entry(number_to_glyph(metric.code), last_glyph);
-	    }
+	    char *nam = new char[strlen(p) + 1];
+	    strcpy(nam, p);
+	    metric.special_device_coding = nam;
+	  }
+	  if (strcmp(nm, "---") == 0) {
+	    last_glyph = number_to_glyph(metric.code);
+	    add_entry(last_glyph, metric);
+	  }
+	  else {
+	    last_glyph = name_to_glyph(nm);
+	    add_entry(last_glyph, metric);
+	    copy_entry(number_to_glyph(metric.code), last_glyph);
 	  }
 	}
-	if (last_glyph == NULL) {
-	  t.error("I didn't seem to find any characters");
-	  return 0;
-	}
       }
-      else {
-	t.error("unrecognised command '%1' "
-		"after 'kernpairs' or 'charset' command",
-		  command);
-	return 0;
+      if (0 == last_glyph) {
+	t.error("no glyphs defined in font description");
+	return false;
       }
     }
-    compact();
+    else {
+      t.error("unrecognized font description directive '%1' (missing"
+	      " 'kernpairs' or 'charset'?)", directive);
+      return false;
+    }
   }
-  if (!is_unicode && !had_charset) {
-    t.error("missing 'charset' command");
-    return 0;
+  compact();
+  t.lineno = 0;
+  if (!saw_name_directive) {
+    t.error("font description 'name' directive missing");
+    return false;
+  }
+  if (!is_unicode && !saw_charset_directive) {
+    t.error("font description 'charset' subsection missing");
+    return false;
   }
   if (space_width == 0) {
+    t.error("font description 'spacewidth' directive missing");
+    // _Don't_ return false; compute a typical one for Western glyphs.
     if (zoom)
-      space_width = scale_round(unitwidth, res, 72 * 3 * sizescale, zoom);
+      space_width = scale_round(unitwidth, res, 72 * 3 * sizescale,
+				zoom);
     else
       space_width = scale_round(unitwidth, res, 72 * 3 * sizescale);
   }
-  return 1;
+  return true;
 }
 
 static struct {
-  const char *command;
+  const char *numeric_directive;
   int *ptr;
 } table[] = {
   { "res", &font::res },
@@ -1028,43 +1062,55 @@ static struct {
   { "sizescale", &font::sizescale },
   };
 
-int font::load_desc()
+// Return file specification of DESC file for selected output device if
+// it can be located and is valid, and a null pointer otherwise.
+const char *font::load_desc()
 {
   int nfonts = 0;
   FILE *fp;
   char *path;
-  if ((fp = open_file("DESC", &path)) == 0) {
-    error("can't find 'DESC' file");
-    return 0;
-  }
+  if ((fp = open_file("DESC", &path)) == 0)
+    return 0 /* nullptr */;
   text_file t(fp, path);
-  t.skip_comments = 1;
-  res = 0;
-  while (t.next()) {
+  while (t.next_line()) {
     char *p = strtok(t.buf, WS);
-    int found = 0;
+    assert(p != 0);
+    bool numeric_directive_found = false;
     unsigned int idx;
-    for (idx = 0; !found && idx < sizeof(table)/sizeof(table[0]); idx++)
-      if (strcmp(table[idx].command, p) == 0)
-	found = 1;
-    if (found) {
+    for (idx = 0; !numeric_directive_found
+		  && idx < sizeof(table) / sizeof(table[0]); idx++)
+      if (strcmp(table[idx].numeric_directive, p) == 0)
+	numeric_directive_found = true;
+    if (numeric_directive_found) {
       char *q = strtok(0, WS);
-      if (!q) {
-	t.error("missing value for command '%1'", p);
-	return 0;
+      if (0 == q) {
+	t.error("missing value for directive '%1'", p);
+	return 0 /* nullptr */;
       }
-      //int *ptr = &(this->*(table[idx-1].ptr));
-      int *ptr = table[idx-1].ptr;
-      if (sscanf(q, "%d", ptr) != 1) {
-	t.error("bad number '%1'", q);
-	return 0;
+      int val;
+      if (sscanf(q, "%d", &val) != 1) {
+	t.error("'%1' directive given invalid number '%2'", p, q);
+	return 0 /* nullptr */;
       }
+      if ((strcmp(p, "res") == 0
+	   || strcmp(p, "hor") == 0
+	   || strcmp(p, "vert") == 0
+	   || strcmp(p, "unitwidth") == 0
+	   || strcmp(p, "paperwidth") == 0
+	   || strcmp(p, "paperlength") == 0
+	   ||  strcmp(p, "sizescale") == 0)
+	  && val < 1) {
+	t.error("expected argument to '%1' directive to be a"
+		" positive number, got '%2'", p, val);
+	return 0 /* nullptr */;
+      }
+      *(table[idx-1].ptr) = val;
     }
     else if (strcmp("family", p) == 0) {
       p = strtok(0, WS);
-      if (!p) {
-	t.error("family command requires an argument");
-	return 0;
+      if (0 == p) {
+	t.error("'family' directive requires an argument");
+	return 0 /* nullptr */;
       }
       char *tem = new char[strlen(p)+1];
       strcpy(tem, p);
@@ -1072,17 +1118,22 @@ int font::load_desc()
     }
     else if (strcmp("fonts", p) == 0) {
       p = strtok(0, WS);
-      if (!p || sscanf(p, "%d", &nfonts) != 1 || nfonts <= 0) {
-	t.error("bad number of fonts '%1'", p);
-	return 0;
+      if (0 == p) {
+	t.error("'fonts' directive requires arguments");
+	return 0 /* nullptr */;
       }
-      font_name_table = (const char **)new char *[nfonts+1]; 
+      if (sscanf(p, "%d", &nfonts) != 1 || nfonts <= 0) {
+	t.error("expected first argument to 'fonts' directive to be a"
+		" non-negative number, got '%1'", p);
+	return 0 /* nullptr */;
+      }
+      font_name_table = (const char **)new char *[nfonts+1];
       for (int i = 0; i < nfonts; i++) {
 	p = strtok(0, WS);
-	while (p == 0) {
-	  if (!t.next()) {
-	    t.error("end of file while reading list of fonts");
-	    return 0;
+	while (0 == p) {
+	  if (!t.next_line()) {
+	    t.error("unexpected end of file while reading font list");
+	    return 0 /* nullptr */;
 	  }
 	  p = strtok(t.buf, WS);
 	}
@@ -1092,48 +1143,61 @@ int font::load_desc()
       }
       p = strtok(0, WS);
       if (p != 0) {
-	t.error("font count does not match number of fonts");
-	return 0;
+	t.error("font count does not match declared number of fonts"
+		" ('%1')", nfonts);
+	return 0 /* nullptr */;
       }
       font_name_table[nfonts] = 0;
     }
     else if (strcmp("papersize", p) == 0) {
-      p = strtok(0, WS);
-      if (!p) {
-	t.error("papersize command requires an argument");
-	return 0;
+      if (0 == res) {
+	t.error("'res' directive must precede 'papersize' in device"
+		" description file");
+	return 0 /* nullptr */;
       }
-      int found_paper = 0;
+      p = strtok(0, WS);
+      if (0 == p) {
+	t.error("'papersize' directive requires an argument");
+	return 0 /* nullptr */;
+      }
+      bool found_paper = false;
+      char *savedp = strdup(p);
+      if (0 == savedp)
+	t.fatal("memory allocation failure while processing 'papersize'"
+		" directive");
       while (p) {
 	double unscaled_paperwidth, unscaled_paperlength;
 	if (scan_papersize(p, &papersize, &unscaled_paperlength,
 			   &unscaled_paperwidth)) {
 	  paperwidth = int(unscaled_paperwidth * res + 0.5);
 	  paperlength = int(unscaled_paperlength * res + 0.5);
-	  found_paper = 1;
+	  found_paper = true;
 	  break;
 	}
 	p = strtok(0, WS);
       }
+      assert(savedp != 0);
       if (!found_paper) {
-	t.error("bad paper size");
-	return 0;
+	t.error("unable to determine a paper format from '%1'", savedp);
+	free(savedp);
+	return 0 /* nullptr */;
       }
+      free(savedp);
     }
     else if (strcmp("unscaled_charwidths", p) == 0)
-      unscaled_charwidths = 1;
+      use_unscaled_charwidths = true;
     else if (strcmp("pass_filenames", p) == 0)
-      pass_filenames = 1;
+      pass_filenames = true;
     else if (strcmp("sizes", p) == 0) {
       int n = 16;
       sizes = new int[n];
       int i = 0;
       for (;;) {
 	p = strtok(0, WS);
-	while (p == 0) {
-	  if (!t.next()) {
+	while (0 == p) {
+	  if (!t.next_line()) {
 	    t.error("list of sizes must be terminated by '0'");
-	    return 0;
+	    return 0 /* nullptr */;
 	  }
 	  p = strtok(t.buf, WS);
 	}
@@ -1147,15 +1211,15 @@ int font::load_desc()
 	    break;
 	  // fall through
 	default:
-	  t.error("bad size range '%1'", p);
-	  return 0;
+	  t.error("invalid size range '%1'", p);
+	  return 0 /* nullptr */;
 	}
 	if (i + 2 > n) {
 	  int *old_sizes = sizes;
 	  sizes = new int[n*2];
 	  memcpy(sizes, old_sizes, n*sizeof(int));
 	  n *= 2;
-	  a_delete old_sizes;
+	  delete[] old_sizes;
 	}
 	sizes[i++] = lower;
 	if (lower == 0)
@@ -1164,7 +1228,7 @@ int font::load_desc()
       }
       if (i == 1) {
 	t.error("must have some sizes");
-	return 0;
+	return 0 /* nullptr */;
       }
     }
     else if (strcmp("styles", p) == 0) {
@@ -1176,7 +1240,7 @@ int font::load_desc()
       int i = 0;
       for (;;) {
 	p = strtok(0, WS);
-	if (p == 0)
+	if (0 == p)
 	  break;
 	// leave room for terminating 0
 	if (i + 1 >= style_table_size) {
@@ -1187,7 +1251,7 @@ int font::load_desc()
 	    style_table[j] = old_style_table[j];
 	  for (; j < style_table_size; j++)
 	    style_table[j] = 0;
-	  a_delete old_style_table;
+	  delete[] old_style_table;
 	}
 	char *tem = new char[strlen(p) + 1];
 	strcpy(tem, p);
@@ -1195,57 +1259,47 @@ int font::load_desc()
       }
     }
     else if (strcmp("tcommand", p) == 0)
-      tcommand = 1;
+      has_tcommand = true;
     else if (strcmp("use_charnames_in_special", p) == 0)
-      use_charnames_in_special = 1;
+      use_charnames_in_special = true;
     else if (strcmp("unicode", p) == 0)
-      is_unicode = 1;
+      is_unicode = true;
     else if (strcmp("image_generator", p) == 0) {
       p = strtok(0, WS);
-      if (!p) {
-	t.error("image_generator command requires an argument");
-	return 0;
+      if (0 == p) {
+	t.error("'image_generator' directive requires an argument");
+	return 0 /* nullptr */;
       }
       image_generator = strsave(p);
     }
     else if (strcmp("charset", p) == 0)
       break;
     else if (unknown_desc_command_handler) {
-      char *command = p;
+      char *directive = p;
       p = strtok(0, "\n");
-      (*unknown_desc_command_handler)(command, trim_arg(p), t.path, t.lineno);
+      (*unknown_desc_command_handler)(directive, trim_arg(p), t.path,
+				      t.lineno);
     }
   }
+  t.lineno = 0;
   if (res == 0) {
-    t.error("missing 'res' command");
-    return 0;
+    t.error("device description file missing 'res' directive");
+    return 0 /* nullptr */;
   }
   if (unitwidth == 0) {
-    t.error("missing 'unitwidth' command");
-    return 0;
+    t.error("device description file missing 'unitwidth' directive");
+    return 0 /* nullptr */;
   }
   if (font_name_table == 0) {
-    t.error("missing 'fonts' command");
-    return 0;
+    t.error("device description file missing 'fonts' directive");
+    return 0 /* nullptr */;
   }
   if (sizes == 0) {
-    t.error("missing 'sizes' command");
-    return 0;
+    t.error("device description file missing 'sizes' directive");
+    return 0 /* nullptr */;
   }
-  if (sizescale < 1) {
-    t.error("bad 'sizescale' value");
-    return 0;
-  }
-  if (hor < 1) {
-    t.error("bad 'hor' value");
-    return 0;
-  }
-  if (vert < 1) {
-    t.error("bad 'vert' value");
-    return 0;
-  }
-  return 1;
-}      
+  return path;
+}
 
 void font::handle_unknown_font_command(const char *, const char *,
 				       const char *, int)
@@ -1259,3 +1313,9 @@ font::set_unknown_desc_command_handler(FONT_COMMAND_HANDLER func)
   unknown_desc_command_handler = func;
   return prev;
 }
+
+// Local Variables:
+// fill-column: 72
+// mode: C++
+// End:
+// vim: set cindent noexpandtab shiftwidth=2 textwidth=72:
